/* code stolen from here and there */

#include "alsa.h"

#include <stdio.h>
#include <stdlib.h>
#include <alsa/asoundlib.h>

static snd_pcm_t *playback_handle;

static const char *device = "default";

int alsa_fd_rd[10];
int alsa_fd_rd_nb;
int alsa_fd_wr[10];
int alsa_fd_wr_nb;

unsigned long frames_per_period;

void init_alsa(void)
{
#define R if (err < 0) { printf("%d:err\n",__LINE__);exit(1); }
  int err;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t period_frames;
  snd_pcm_uframes_t buffer_frames;
  unsigned int rate = 44100;
  err = snd_pcm_open(&playback_handle, device, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK); R;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(playback_handle, hw_params); R;
  err = snd_pcm_hw_params_set_access(playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); R;
  err = snd_pcm_hw_params_set_format(playback_handle, hw_params, SND_PCM_FORMAT_S16_LE); R;
  err = snd_pcm_hw_params_set_channels(playback_handle, hw_params, 2); R;
  err = snd_pcm_hw_params_set_rate_near(playback_handle, hw_params, &rate, 0); R;
  if (rate != 44100) {printf("bad rate\n"); exit(1);}
#if 0
  period_frames = 512 / 4;
#endif
  period_frames = 1024;
  buffer_frames = period_frames * 100;
  err = snd_pcm_hw_params_set_period_size_near(playback_handle, hw_params, &period_frames, 0); R;
  err = snd_pcm_hw_params_set_buffer_size_near(playback_handle, hw_params, &buffer_frames); R;
  err = snd_pcm_hw_params(playback_handle, hw_params); R;
  printf("period_frame %lu buffer_frames %lu\n", period_frames, buffer_frames);
  frames_per_period = period_frames;
  if (period_frames < 512/4) { printf("period size too small\n"); exit(1); }
  if (buffer_frames < period_frames * 2)
    { printf("buffer frames too small\n"); exit(1); }
  snd_pcm_sw_params_current(playback_handle, sw_params);
{
  snd_pcm_uframes_t v;
  snd_pcm_sw_params_get_avail_min(sw_params, &v);
  if (v != period_frames) { printf("avail_min wrong\n"); exit(1); }
#if 0
  if (v != 512/4) {
    err = snd_pcm_sw_params_set_avail_min(playback_handle, sw_params, 512/4); R;
    err = snd_pcm_sw_params(playback_handle, sw_params); R;
    snd_pcm_sw_params_current(playback_handle, sw_params);
    snd_pcm_sw_params_get_avail_min(sw_params, &v);
    if (v != 512/4 && (v & (512/4-1))) {
      snd_pcm_uframes_t nv = v + 128;
      err = snd_pcm_sw_params_set_avail_min(playback_handle, sw_params, nv); R;
      err = snd_pcm_sw_params(playback_handle, sw_params); R;
      snd_pcm_sw_params_current(playback_handle, sw_params);
      snd_pcm_sw_params_get_avail_min(sw_params, &v);
      if (v != nv) goto amr;
      if (buffer_frames < 1024/2 + 128)
        { printf("buffer_frames not ok\n"); exit(1); }
    }
    if (v < 512/4) {
amr:  printf("can't set avail min\n"); exit(1);
    }
  }
#endif
  printf("avail_min %lu\n", v);
}
{
  snd_pcm_uframes_t v;
  snd_pcm_sw_params_get_start_threshold(sw_params, &v);
  printf("start_threshold %lu\n", v);
  snd_pcm_sw_params_get_stop_threshold(sw_params, &v);
  printf("stop_threshold %lu\n", v);
}
  {
    int i;
    struct pollfd desc[10];
    int ndesc = snd_pcm_poll_descriptors_count(playback_handle);
    if (ndesc > 10) { printf("ALSA weirdness 1 (ndesc %d)\n",ndesc); exit(1); }
    if (snd_pcm_poll_descriptors(playback_handle, desc, ndesc) != ndesc)
      {printf("ALSA weirdness 2\n"); exit(1);}
    for (i = 0; i < ndesc; i++) {
      if (desc[i].events & POLLOUT) {
        alsa_fd_wr[alsa_fd_wr_nb] = desc[i].fd;
        alsa_fd_wr_nb++;
      }
      if (desc[i].events & POLLIN) {
        alsa_fd_rd[alsa_fd_rd_nb] = desc[i].fd;
        alsa_fd_rd_nb++;
      }
    }
    if (!alsa_fd_wr_nb && !alsa_fd_rd_nb) { printf("ALSA weirdness 3\n"); exit(1); }
  }
  err = snd_pcm_prepare(playback_handle); R;
{
  char *b = malloc(4 * period_frames * 2); if (!b) R;
  memset(b, 0, 4 * period_frames * 2);
  play_alsa(b, 4 * period_frames * 2);
  free(b);
}
}

static int xrun_recovery(snd_pcm_t *playback_handle, int err)
{
  printf("\nxrun_recovery err is %d EPIPE is %d ESTRPIPE is %d\n", err, -EPIPE, -ESTRPIPE);
        if (err == -EPIPE) {    /* under-run */
                err = snd_pcm_prepare(playback_handle);
                if (err < 0) {
                        printf("Can't recovery from underrun, prepare failed: %s\n", snd_strerror(err));
	                return err;
		}
		return 0;
        } else if (err == -ESTRPIPE) {
                while ((err = snd_pcm_resume(playback_handle)) == -EAGAIN)
                        sleep(1);       /* wait until the suspend flag is released */
                if (err < 0) {
                        err = snd_pcm_prepare(playback_handle);
                        if (err < 0) {
                                printf("Can't recovery from suspend, prepare failed: %s\n", snd_strerror(err));
				return err;
			}
                }
                return 0;
        }
        return err;
}

static int last_ev;
void play_alsa(const char *b, int l)
{
  int err;
  int nb_try = 10;
//printf("play %d (avail %lu)\n", l, snd_pcm_avail_update(playback_handle));
again:
  if ((err = snd_pcm_writei (playback_handle, b, l/4)) != l/4) {
    if (err == -EAGAIN) {
      printf("AGAIN (state %d last_ev %d av %lu av_up %lu)\n",
             snd_pcm_state(playback_handle), last_ev,
             snd_pcm_avail(playback_handle),
             snd_pcm_avail_update(playback_handle));
      goto again;
    }
    if (err >= 0) printf("weird...\n");
    if (err > 0) { b += err*4; l -= err; goto again; }
    if ((err = xrun_recovery(playback_handle, err))) {
      fprintf(stderr, "alsa out err\n");
      exit(1);
    }
    nb_try--; if (nb_try) goto again;
  }
}

void end_alsa(void)
{
  snd_pcm_uframes_t total;
  snd_pcm_uframes_t avail;
  snd_pcm_hw_params_t *hw_params;
  int err;

  snd_pcm_hw_params_alloca(&hw_params);
  if ((err=snd_pcm_hw_params_current(playback_handle, hw_params))<0) R;
  if ((err=snd_pcm_hw_params_get_buffer_size(hw_params, &total))<0) R;
  do {
    sleep(1);
    avail = snd_pcm_avail(playback_handle);
  } while (snd_pcm_state(playback_handle) == SND_PCM_STATE_RUNNING &&
           avail < total);

  snd_pcm_drain(playback_handle);
  snd_pcm_close(playback_handle);
}

int alsa_can_play(void *_rd, void *_wr)
{
  fd_set *rd = _rd;
  fd_set *wr = _wr;
  struct pollfd p[10];
  int pos = 0;
  int i;
  unsigned short ev = 0;
  int err;
  for (i = 0; i < alsa_fd_rd_nb; i++) {
    p[pos].fd = alsa_fd_rd[i];
    p[pos].events = POLLIN;
    p[pos].revents = 0;
    if (FD_ISSET(alsa_fd_rd[i], rd)) p[pos].revents = POLLIN;
    pos++;
  }
  for (i = 0; i < alsa_fd_wr_nb; i++) {
    p[pos].fd = alsa_fd_wr[i];
    p[pos].events = POLLOUT;
    p[pos].revents = 0;
    if (FD_ISSET(alsa_fd_wr[i], wr)) p[pos].revents = POLLOUT;
    pos++;
  }
  err=snd_pcm_poll_descriptors_revents(playback_handle, p, pos, &ev); R;
  last_ev = ev;
  if (ev & POLLOUT) return 1;
  return 0;
}
